import java.io.*;
import java.util.*;

abstract class Polys {

    static final double ROOT2 = Math.sqrt(2);
    public static Hashtable pdict;
    static float camfactor = 8;
    private static PrintWriter output;
    public Hashtable edict, vdict, fdict;
    double volume;
    String edgecolor="Gray", facecolor, vertcolor="Gray";
    float edgediam = (float)0.02, vertdiam = (float)0.02;
    boolean byx = true, byy = true, byz = true;
    boolean showedges = true, showfaces = false, showverts = true;
    Quaternion rotx,roty,rotz,invrotx,invroty,invrotz;

    Polys(){
	    //concentric hierarchy vertices common to all polys
            pdict = new Hashtable(26);

            pdict.put("A",new Vertex(new double[]{1,0,0,0}));
            pdict.put("B",new Vertex(new double[]{0,1,0,0}));
            pdict.put("C",new Vertex(new double[]{0,0,1,0}));
            pdict.put("D",new Vertex(new double[]{0,0,0,1}));

            pdict.put("E",getp("A").inv());
            pdict.put("F",getp("B").inv());
            pdict.put("G",getp("C").inv());
            pdict.put("H",getp("D").inv());

	    addvert("I","A","B");
	    addvert("J","A","C");
	    addvert("K","A","D");
	    addvert("L","B","C");
	    addvert("M","B","D");
	    addvert("N","C","D");

	    addvert("O","I","J");
	    addvert("P","I","K");
	    addvert("Q","I","L");
	    addvert("R","I","M");
	    addvert("S","N","J");
	    addvert("T","N","K");

	    addvert("U","N","L");
	    addvert("V","N","M");
	    addvert("W","J","L");
	    addvert("X","L","M");
	    addvert("Y","M","K");
	    addvert("Z","K","J");

	    setrotate(0); // default

     }

     // used by constructor
     private Vertex getv(String ckey){
	return (Vertex)vdict.get(ckey);
     }

     private static Vertex getp(String ckey){
	return (Vertex)pdict.get(ckey);
     }

     private static void addvert(String ckey, String v1, String v2){
            pdict.put(ckey,getp(v1).add(getp(v2)));
     }

     // transfer verts specific to poly from pdict to vdict (constructor)
     protected void addverts(String vlist){
	for (int i=0;i<vlist.length() ;i++ )
	    addvert((new Character(vlist.charAt(i))).toString());
     }

     private void addvert(String ckey){
            vdict.put(ckey,getp(ckey));
     }

     // store edges specific to poly in edict (constructor)
     protected void addedges(String[] elist){
	Character a,b;
	for (int i=0;i<elist.length;i++){
	    a=new Character(elist[i].charAt(0));
	    b=new Character(elist[i].charAt(1));
      	    addedge(elist[i],a.toString(),b.toString());
	}
     }

     private void addedge(String ckey, String e1, String e2){
            edict.put(ckey,new Edge(e1,e2));
     }

     // store faces specific to poly in fdict (constructor)
     protected void addfaces(String[] flist){
	String face;
	String[] vertlist;
	for (int i=0;i<flist.length;i++){
	    face = flist[i];
	    vertlist = new String[face.length()];
	    for (int j=0;j<face.length();j++)
	    {
	        vertlist[j]=new Character(face.charAt(j)).toString();
	    }
            addface(flist[i],vertlist);
	}
     }

     private void addface(String ckey, String[] vertlist){
            fdict.put(ckey,new Face(vertlist));
     }

     public void writexyz(){
	 Edge edge;
 	 Enumeration enum = edict.elements();
         while (enum.hasMoreElements()){
            edge = (Edge)enum.nextElement();
	    edge.xyzout();
         }
     }

     public void writepovray(){
	if ( showedges ) writeEdges();
	if ( showfaces ) writeFaces();
	if ( showverts ) writeVerts();
     }

     public void writeEdges(){
 	Edge edge;
	Enumeration enum = edict.elements();
        while (enum.hasMoreElements()){
            edge = (Edge)enum.nextElement();
	    edge.writeEdge();
	}
     }

     public void writeFaces(){
 	Face face;
	Enumeration enum = fdict.elements();
        while (enum.hasMoreElements()){
            face = (Face)enum.nextElement();
	    face.writeFace();
	}
     }

     public void writeVerts(){
 	Vertex vert;
	Enumeration enum = vdict.elements();
        while (enum.hasMoreElements()){
            vert = (Vertex)enum.nextElement();
	    if (!vert.tickoff) vert.writeVert();
	}
	enum = vdict.elements();
        while (enum.hasMoreElements()){
            vert = (Vertex)enum.nextElement();
	    vert.tickoff = false;
	}
     }

     public void setVerts(boolean value){
	Enumeration enum = vdict.elements();
        while (enum.hasMoreElements()){
            ((Vertex)enum.nextElement()).showvert = value;
	}
     }

     public void setVert(String ckey, boolean value){
	getv(ckey).showvert = value;
     }

     public void setEdges(boolean value){
	Enumeration enum = edict.elements();
        while (enum.hasMoreElements()){
            ((Edge)enum.nextElement()).showedge = value;
	}
     }

     public void setEdge(String ckey, boolean value){
	((Edge)edict.get(ckey)).showedge = value;
     }

     public void setFaces(boolean value){
	Enumeration enum = fdict.elements();
        while (enum.hasMoreElements()){
            ((Face)enum.nextElement()).showface = value;
	}
     }

     public void setFace(String ckey, boolean value){
	((Face)fdict.get(ckey)).showface = value;
     }

     protected void setrotate(double deg){
            rotx = new Quaternion("X",deg); // rotate 1 degree around X
            roty = new Quaternion("Y",deg); // rotate 1 degree around Y
            rotz = new Quaternion("Z",deg); // rotate 1 degree around Z

            // I need the inverse of each of the above -- Quaternions
            // know how to return their own inverses
            invrotx = rotx.inv();
            invroty = roty.inv();
	    invrotz = rotz.inv();
     }

     protected void rotate(){
	Vertex vert;
	String thiskey;
        Enumeration vnum = vdict.keys();
        while (vnum.hasMoreElements()){
            // I'm getting the vector part after rotating each
            // vertex i.e. c = (q c q^-1).vector

	    thiskey = (String)vnum.nextElement();
	    vert = getv(thiskey);

            vert  = new Vertex(rotx.mult(vert.xyz).mult(invrotx).vector);
            vert  = new Vertex(roty.mult(vert.xyz).mult(invroty).vector);
            vert  = new Vertex(rotz.mult(vert.xyz).mult(invrotz).vector);

	    vdict.put(thiskey,vert);
        }
     }

    class Vertex {

       double[] tuple;
       double[] xyz;
       boolean tickoff;
       boolean showvert = true;

       Vertex(double[] coords){

          tuple = new double[4];
          xyz = new double[3];

          if ( 4==coords.length){
              for (int i=0;i<4;i++) tuple[i] = coords[i];
              quad2xyz();
          } else {
              for (int i=0;i<3;i++) xyz[i] = coords[i];
	      xyz2quad();
          }
       }

       public Vertex inv(){
          return simplify(new Vertex(new double[]
                                 {-tuple[0],
                                  -tuple[1],
                                  -tuple[2],
                                  -tuple[3]}
                          )
                     );
       }

       public Vertex add(Vertex v){
          return simplify(new Vertex(new double[]
                                 {tuple[0]+v.tuple[0],
                                  tuple[1]+v.tuple[1],
                                  tuple[2]+v.tuple[2],
                                  tuple[3]+v.tuple[3]}
                          )
                      );
       }

       public Vertex mult(double s){
          return (new Vertex(new double[]
                             {s*tuple[0],
                              s*tuple[1],
                              s*tuple[2],
                              s*tuple[3]}
                          )
                      );
       }

       private Vertex simplify(Vertex v){
          double minval = v.tuple[0];
          for (int i=1;i<4;i++) minval = minval<v.tuple[i]?minval:v.tuple[i];
          for (int i=0;i<4;i++) v.tuple[i] -= minval;
          return v;
       }

       private void simplify(){
          double minval = tuple[0];
          for (int i=1;i<4;i++) minval = minval<tuple[i]?minval:tuple[i];
          for (int i=0;i<4;i++) tuple[i] -= minval;
       }

       private void quad2xyz(){
            xyz[0] = 1/ROOT2 * (tuple[0] - tuple[1] - tuple[2] + tuple[3]);
            xyz[1] = 1/ROOT2 * (tuple[0] - tuple[1] + tuple[2] - tuple[3]);
            xyz[2] = 1/ROOT2 * (tuple[0] + tuple[1] - tuple[2] - tuple[3]);
       }

       private void xyz2quad(){
            tuple[0] = 1/ROOT2 *
	               ((xyz[0]>=0? xyz[0]: 0)+(xyz[1]>=0? xyz[1]: 0)+(xyz[2]>=0? xyz[2]: 0));
            tuple[1] = 1/ROOT2 *
	               ((xyz[0]>=0? 0:-xyz[0])+(xyz[1]>=0? 0:-xyz[1])+(xyz[2]>=0? xyz[2]: 0));
            tuple[2] = 1/ROOT2 *
	               ((xyz[0]>=0? 0:-xyz[0])+(xyz[1]>=0? xyz[1]: 0)+(xyz[2]>=0? 0:-xyz[2]));
            tuple[3] = 1/ROOT2 *
	               ((xyz[0]>=0? xyz[0]: 0)+(xyz[1]>=0? 0:-xyz[1])+(xyz[2]>=0? 0:-xyz[2]));
            simplify();
       }

       public String quadout(){
            return "<"+tuple[0]+", "
                      +tuple[1]+", "
                      +tuple[2]+", "
                      +tuple[3]+">";
       }

       public String xyzout(){
	    // write vertex as <x,y,z> (String)
            return "<"+xyz[0]+", "
                      +xyz[1]+", "
                      +xyz[2]+">";
       }

       public void writeVert(){
	    // write vertex as sphere{} in povray format
	    if ( showvert ){
                output.println("sphere{ <"+xyz[0]+", "+xyz[1]+", "+xyz[2]+">," +
	                    vertdiam + " pigment {color " + vertcolor + "} no_shadow}");
    	        tickoff = true;
	    }
       }

    }

    // Edge class stores pair of vertexes as String[]
    class Edge {

       String[] pair;
       boolean showedge=true;

       Edge(String v1, String v2){
           pair = new String[]{v1,v2};
       }

       public void xyzout(){
           System.out.println("From: "+getv(pair[0]).xyzout()+" to "
                                      +getv(pair[1]).xyzout());
       }

       public void writeEdge(){
	    // write edge as cylinder{} in povray format
	    if ( showedge ){
                output.println(
                    "cylinder{"+ getv(pair[0]).xyzout() + ","
 		               + getv(pair[1]).xyzout() + ","
                               + edgediam + " pigment {color " + edgecolor + "} no_shadow}");
	    }
       }


    }

    // Face class stores list of vertexes as String[]
    class Face {

       String[] verts;
       boolean showface=true;

       Face(String[] f){
           verts = f;
       }

       public void writeFace(){
	    // write face as polygon{} in povray format
	    if ( showface ){
                int nbpoints = verts.length + 1;
                output.println("polygon{"+ nbpoints +",");
	        for ( int i=0;i<verts.length ;i++)
	        {
        	    output.println(getv(verts[i]).xyzout() + ",");
	        }
  	        output.println(getv(verts[0]).xyzout());
	        output.println(" pigment {color " + facecolor + "}}");
	    }
       }
    }

    public static void openfile(String filename){
        try{
            output =
              new PrintWriter(
              new BufferedWriter(
              new FileWriter(filename)));
        } catch(Exception e) {System.out.println("Couldn't create file!");}
	writeHeader();
    }

    public static void closefile(){
	output.close();
    }

    public static void writeHeader(){
          output.println("//POV-Ray script");
          output.println("#version 3.1;");
          output.println("global_settings { assumed_gamma 2.2 }");
          output.println("#include \"colors.inc\"");
          output.println("");
          output.println("#declare Cam_factor = " + camfactor + ";");
          output.println("#declare Camera_X = 1 * Cam_factor;");
          output.println("#declare Camera_Y = 0.5 * Cam_factor;");
          output.println("#declare Camera_Z = -0.9 * Cam_factor;");
          output.println("camera { location  <Camera_X, Camera_Y, Camera_Z>");
          output.println("            up        <0, 1.0,  0>    right     <-4/3, 0,  0>");
          output.println("            direction <0, 0,  3>      look_at   <0, 0, 0> ");
          output.println("            rotate <0,0,0>}");
          output.println("");
          output.println("light_source { <Camera_X - 2, Camera_Y + 5 , Camera_Z + 5> color White }");
          output.println("light_source { <Camera_X - 2, Camera_Y + 5 , Camera_Z - 3> color White }");
          output.println("");
          output.println("// Background:");
          output.println("background {color White}");
    }
}

